#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-12-22
#
# Licensed under Apache License, Version 2.0.
#

# The alias of the current package: the one Context object was initialized with.
CURRENT = "."

# The top level package: the top-most found by Context while traversing '..'.
ROOT = "/"

# Which file to look for if the directory name is passed instead of the config
# filename.
DEFAULT_PACKAGE_CONFIG = "partcad.yaml"
