#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2023-08-19
#
# Licensed under Apache License, Version 2.0.
#

from .. import partcad as pc


class Scene(pc.Assembly):
    def __init__(self):
        super().__init__()
