#
# OpenVMP, 2023
#
# Author: Roman Kuzmenko
# Created: 2024-02-19
#
# Licensed under Apache License, Version 2.0.
#

from . import factory


class ShapeFactory(factory.Factory):
    def __init__(self) -> None:
        super.__init__()
